/////////////////////////////////////////////////////////////////////////////////////////////
//	ProtocolStack fr Arduino Intervall-Timer mit Bluetooth-Modul
// 
//       18.10.2024	                             Erwin Lotter
/////////////////////////////////////////////////////////////////////////////////////////////
#ifndef protocol_h_
#define protocol_h_


#define PROTOCOL_STACK_SIZE   500     // reduce if Arduino gets out of SDRam

#if defined(T_SENSOR_AVAILABLE) && defined(AHT20_AVAILABLE)
#define TWO_TEMP
#endif

/*
Jeder Eintrag in ProtocolStack enthlt den Frameindex der Aufnahme seit dem Einschalten
und fr jeden Sensor einen Temperturwert in der Reihenfolge T_SENSOR_AVAILABLE, AHT20_AVAILABLE.
Jeder der Eintrge belegt ein int16_t, es sei denn, er wird komprimiert.

Sobald mindestens ein Wert komprimiert wird, wird ein komprimiertes Bytes (K-Byte)
gespeichert, das immer daran zu erkennen ist, dass sein MSB gesetzt ist.

- Wenn der Frameindex gleich dem vorherigen plus eins ist, entfllt sein Eintrag
  und das erste Bytes ist ein K-Byte. 
  Andernfalls ist das erste Wort der Frameindex (max. 0x7FFF). 
  Wenn diesem kein K-Byte folgt, ist es der erste, unkomprimierte Temperturwert.
  Wenn dann nochmals kein K-Byte folgt, ist das der zweite unkomprimierte Temperturwert.
  
- Wenn sich ein Temperturwert um maximal +/-d von seinem Vorgnger unterscheidet,
  wird die Differenz in n Bits des K-Bytes geschrieben:
   Fr nur einen Temperturwert sind das Bit 6-0 und d = 63.
   Fr zwei Temperturwerte erhlt T1 Bit 6-3 mit d = 7 und T2 erhlt Bit 2-0 mit d = 3.
  Der Wert -d-1 zeigt jeweils an, dass ein voller Eintrag notwendig war, der dann
  dem K-Byte folgt. Zur einfacheren und  vorzeichenlosen Verarbeitung wird zur Differenz
  d+1 addiert, so dass dann 0 statt (-d-1) einen berlauf der Differenz anzeigt.
*/

#if defined(TWO_TEMP)
const int dT1_shift = 3;
const int dT1_limit = 8;
const int dT2_shift = 0;
const int dT2_limit = 4;
const int dT2_mask  = (dT2_limit * 2 -1);
#else
const int dT1_shift = 0;
const int dT1_limit = 64;
#endif
const int dT1_mask  = (dT1_limit * 2 -1);

class ProtocolStack {
  public:
	    ProtocolStack() { Reset(); }
    void    Reset();
    void    Stop() { active = false; }
    void    Continue() { active = true; }
    bool    IsActive() const { return active; }
    int	    GetCount() const { return cnt; }
    bool    IsFull() const { return nextWritePos > pbf + maxEntryOfs; }

    bool    IsSending() const { return sending; }
    bool    StartSending();
    bool    ReadNextLine();

    #if defined(TWO_TEMP)
    bool    Add(int frame, int T1, int T2);
    int	    GetRead_T2() const  {return readT2;}
    enum { maxEntrySize = 6 }; // max 6 Bytes
    int	    prevT2;	// previuos T2
    int	    readT2;	// gelesene T1
    #else
    bool    Add(int frame, int T1);
    enum { maxEntrySize = 4 };
    #endif
    int	    GetRead_I() const   {return readIndex;}
    int	    GetRead_T1() const  {return readT1;}

    enum { maxEntryOfs = PROTOCOL_STACK_SIZE - maxEntrySize };  // noch Platz fr einen vollen Eintrag
  protected:
    uint8_t ReadByte();
    int	    AppendReadByte(int8_t k_byte);
    int     ReadWord();
    void    WriteWord(int w);
    void    WriteByte(int8_t b);
    int	    ExtractT1(int8_t k_byte);
    int	    ExtractT2(int8_t k_byte);
    void    DecompressT1(int8_t k_byte);
    void    DecompressT2(int8_t k_byte);

  protected:
    int8_t  pbf[PROTOCOL_STACK_SIZE];
    int8_t* nextWritePos;
    int8_t* nextReadPos;
    int	    cnt;	// index of next entry to add
    int	    prevIndex;	// previuos index
    int	    prevT1;	// previuos T1
    int	    readIndex;	// gelesener Index
    int	    readT1;	// gelesene T1
    bool    active;
    bool    sending;
};


//
// Inhalt lschen, Schreibposition auf 0, aktiv setzen.
//
inline
void ProtocolStack::Reset()
{
  sending = false;
  cnt = 0;
  nextWritePos = pbf;	  // Schreibposition absolut
  prevT1 = 0;		// zuvor geschriebenes T1
  #if defined(TWO_TEMP)
  prevT2 = 0;
  #endif
  active = true;
}


//
// Asugabe init., false wenn nix da.
//
inline
bool ProtocolStack::StartSending()
{
  sending = GetCount() > 0;
  nextReadPos = pbf;	  // Leseposition absolut
  readIndex = 0;	// zuvor gelesener Index 
  readT1 = 0;		// zuvor gelesene T1
  #if defined(TWO_TEMP)
  readT2 = 0;		// zuvor gelesene T2
  #endif
  return sending;
}


//
// Byte entnehmen, Leseadresse weiter
//
inline
uint8_t ProtocolStack::ReadByte()
{
  return *nextReadPos++;
}

//
// k_byte und das nchste Byte zu einem signed 15-Bit Wert kombinieren
//  und auf 16 Bit erweitern. 
// Leseadresse weiter.
//
inline
int ProtocolStack::AppendReadByte(int8_t k_byte)
{
  return (int16_t(uint16_t(k_byte) << 9) >> 1) | ReadByte();
}

//
// Signed 15-Bit Wert entnehmen, auf 16 Bit erweitern.
// Leseadresse weiter.
//
inline
int ProtocolStack::ReadWord()
{
  uint8_t r = ReadByte();
  return AppendReadByte(r);
}

//
// Byte eintragen, Schreibadresse weiter
//
void bluePrint_P(PGM_P s);
void bluePrint(int);
inline
void ProtocolStack::WriteByte(int8_t b)
{
  *nextWritePos++ = b;
//bluePrint(b); bluePrint_P(PSTR(" "));
}

//
// 16-Bit Wert eintragen, Schreibadresse weiter
//
inline
void ProtocolStack::WriteWord(int w)
{
  WriteByte(uint8_t(w >> 8) & 0x7F); // Index komplett schreiben, HiByte zuerst,
  WriteByte(uint8_t(w));	     //  damit das hchste/erste Bit sicher 0 bleibt
}


//
// T1 aus K-Byte entnehmen
//
inline
int ProtocolStack::ExtractT1(int8_t k_byte)
{
  return (k_byte >> dT1_shift) & dT1_mask;
}


//
// T2 aus K-Byte entnehmen
//
inline
int ProtocolStack::ExtractT2(int8_t k_byte)
{
  return (k_byte >> dT2_shift) & dT2_mask;
}


//
// T1 aus K-Byte entnehmen oder unkomprimiert lesen 
//  und in readT1 eintragen.
//
inline
void ProtocolStack::DecompressT1(int8_t k_byte)
{
  int dT1 = ExtractT1(k_byte);
  if (dT1 == 0) {				// T1 ist nicht komprimiert:
    readT1 = ReadWord(); // lesen
  }
  else {
    readT1 += dT1 - dT1_limit;
  }
}

//
// T2 aus K-Byte entnehmen oder unkomprimiert lesen
//  und in readT2 eintragen.
//
inline
void ProtocolStack::DecompressT2(int8_t k_byte)
{
  #if defined(TWO_TEMP)
  int dT2 = ExtractT2(k_byte);
  if (dT2 == 0) {				// T2 ist nicht komprimiert:
    readT2 = ReadWord(); // lesen
  }
  else {
    readT2 += dT2 - dT2_limit;
  }
  #endif
}

////////////////////////////////////////////////////////////////////////
// Nchste Zeile des Protokolls lesen, Position weitersetzen
//  und 'sending' ggf. abschalten
//
inline
bool ProtocolStack::ReadNextLine()
{
  sending = (nextReadPos < nextWritePos);   // noch was da?
  if (sending) {
    int8_t k_byte = ReadByte();
    readIndex++;				// Frameindex 1 weiter
    if (k_byte >= 0) {			        // Frameindex ist nicht komprimiert:
      readIndex = (k_byte << 8) | ReadByte();	// lesen
      k_byte = ReadByte();			// nchstes Byte prfen
      if (k_byte >= 0) {		        // T1 ist nicht komprimiert:
	readT1 = AppendReadByte(k_byte);	//  noch ein Byte dazu
	#if defined(TWO_TEMP)
	k_byte = ReadByte();			// nchstes Byte prfen
	if (k_byte >= 0) {			// T2 ist nicht komprimiert:
	  readT2 = AppendReadByte(k_byte);	//  noch ein Byte dazu
	}
	else {
	  readT2 += ExtractT2(k_byte) - dT2_limit;
	}
	#endif
      }
      else {					// T1 ist komprimiert:
	readT1 += ExtractT1(k_byte) - dT1_limit;
	DecompressT2(k_byte);			// T2 ist evtl. komprimiert
      }
    }
    else {	        		        // Frameindex ist komprimiert:
      DecompressT1(k_byte);			// T1 ist evtl. komprimiert
      DecompressT2(k_byte);			// T2 ist evtl. komprimiert
    }
  }
  return sending;
}


//
// Wert in Protokoll eintragen
//
inline
bool ProtocolStack::Add(int frame, int T1
    #if defined(TWO_TEMP)
    , int T2
    #endif
) {
  if (!IsActive() || IsFull()) {
    return false;
  }
  uint8_t k_byte = 0x80;		// K-Byte erstellen:
  int dT1 = T1 - prevT1;		// T1-Diff + 8 = 1...15, wenn in Bit 6-3 unterzubringen
  if (abs(dT1) < dT1_limit) {		// passt:
    k_byte |= ((dT1 + dT1_limit) << dT1_shift);//  ins K-Byte eintragen
  }
  prevT1 = T1;
  #if defined(TWO_TEMP)
  int dT2 = T2 - prevT2;		// T2-Diff + 4 = 1...7, wenn in Bit 2-0 unterzubringen
  if (abs(dT2) < dT2_limit) {		// passt:
    k_byte |= ((dT2 + dT2_limit) << dT2_shift);//  ins K-Byte eintragen
  }
  prevT2 = T2;
  #endif
  prevIndex++;				// Folgeindex
  if (frame != prevIndex) {		// nicht letzter Index + 1:
    prevIndex = frame;
    WriteWord(frame);			//  Index komplett schreiben
    if (abs(dT1) >= dT1_limit) {	// T1 passt nicht:
      WriteWord(T1);			//  komplett schreiben
      #if defined(TWO_TEMP)
      if (abs(dT2) >= dT2_limit) {	// T2 passt auch nicht:
	WriteWord(T2);			//  komplett schreiben
      }
      else {				// sonst jetzt das K-Byte
	WriteByte(k_byte);	// K-Byte steht in der Mitte
      }
      #endif
    }
    else {				// sonst jetzt das K-Byte
      WriteByte(k_byte);		// K-Byte steht in der Mitte
      #if defined(TWO_TEMP)
      if (abs(dT2) >= dT2_limit) {	// T2 passt auch nicht:
	WriteWord(T2);			//  komplett schreiben
      }
      #endif
    }
  }
  else {				// Index ist 1 weiter:
    WriteByte(k_byte);			// K-Byte steht vorn
    if (abs(dT1) >= dT1_limit) {	// T1 passt nicht:
      WriteWord(T1);			//  komplett schreiben
    }
    #if defined(TWO_TEMP)
    if (abs(dT2) >= dT2_limit) {	// T2 passt nicht:
      WriteWord(T2);			//  komplett schreiben
    }
    #endif
  }
  cnt++;
//bluePrint_P(PSTR("\n"));
  return true;
}  
  
#endif
